part of http_router;

class HttpRouter extends StreamConsumer<HttpRequest> {

  /**
   * Mapping of hosts, to methods, to routes
   */
  Dictionary _routes;

  /**
   * The fallback route handler when no match is found
   */
  HttpRouteHandler _fallback;

  /**
   * The internal service manager reference
   */
  ServiceManager _serviceManager;

  /**
   * Constructor
   *
   * @param ServiceManager serviceManager - The service manager
   * @param Map routes                    - Map of hosts to methods to routes
   * @param HttpRouteHandler              - The fallback route handler
   */
  HttpRouter(this._serviceManager, Map routes, this._fallback) {
    var root = new Dictionary.fromMap(routes);
    this._routes = root.fold(new Dictionary(), (memo, v, k) {
      var tmp = new Dictionary.fromMap(v);
      memo[this._keyToRegExp(k)] = tmp.fold(new Dictionary(), (memo, v, k) {
        var tmp = new Dictionary.fromMap(v);
        memo[k] = tmp.fold(new Dictionary(), (memo, v, k) {
          memo[this._keyToRegExp(k)] = v;
          return memo;
        });
        return memo;
      });
      return memo;
    });
  }

  /**
   * Required for implementation of `StreamConsumer`
   *
   * @param Stream<HttpRequest> requests - The stream of http requests
   * @return Future<bool>                - The completion future
   */
  Future<bool> addStream(Stream<HttpRequest> requests) {
    var completer = new Completer<bool>();
    requests.listen((HttpRequest request) {

      var route = this._getHostHeader(request).flatMap((host) {
        return this._getHostMethods(host);
      }).flatMap((methods) {
        return this._getMethod(request, methods);
      }).flatMap((routes) {
        return this._getRoute(request, routes);
      });

      this._handleRoute(request, route);
    });
    return completer.future;
  }

  /**
   * Helper method for getting the host from the http request
   *
   * @param HttpRequest request - The http request
   * @return Option<String>     - The optional host
   */
  Option<String> _getHostHeader(HttpRequest request) {
    var hosts = request.headers["host"];
    if (hosts == null) {
      return new None();
    } else {
      return new Some(hosts[0]);
    }
  }

  /**
   * Helper method for getting the methods of a host
   *
   * @param String host         - The host to get the methods of
   * @return Option<Dictionary> - The optional methods of the host
   */
  Option<Dictionary> _getHostMethods(String host) {
    return this._routes.findValue((value, key) {
      return key.hasMatch(host);
    });
  }

  /**
   * Helper method for getting the specific method from host methods
   *
   * @param HttpRequest request - The http request to derive the method from
   * @param Dictionary methods  - The map of host methods
   * @return Option<Dictionary> - The optional method dictionary
   */
  Option<Dictionary> _getMethod(HttpRequest request, Dictionary methods) {
    var method = request.method;
    return methods.get(method);
  }

  /**
   * Helper method for getting the correct route from a method
   *
   * @param HttpRequest request      - The http request
   * @param Dictionary routes        - The map of routes per a method
   * @return Option<HttpRouteHander> - The optional route handler
   */
  Option<HttpRouteHandler> _getRoute(HttpRequest request, Dictionary routes) {
    var path = request.uri.path;
    return routes.findValue((value, key) {
      return key.hasMatch(path);
    });
  }

  /**
   * Helper method for handling a resolved route
   *
   * @param HttpRequest request              - The http request
   * @param Option<HttpRouteHandler> handler - The resolved route
   * @return void
   */
  void _handleRoute(HttpRequest request, Option<HttpRouteHandler> handler) {
    if (handler.nonEmpty()) {
      handler.get().handler(this._serviceManager, request);
    } else {
      this._fallback.handler(this._serviceManager, request);
    }
  }

  /**
   * Helper method for converting a key string to a regexp
   *
   * @param String key - The key to convert
   * @return RegExp    - The computed regex
   */
  RegExp _keyToRegExp(String key) {
    if (key.startsWith('r:')) {
      return new RegExp(key.substring(2));
    } else {
      return new RegExp('^' + key + '\$');
    }
  }

  /**
   * Required for implementation of `StreamConsumer`
   *
   * @return Future<bool>
   */
  Future<bool> close() {
    var completer = new Completer<bool>();
    return completer.future;
  }

}